/*
 *  Copyright (c) 2007 Jiri Benc <jbenc@upir.cz>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <glib.h>
#include <glib/gprintf.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#ifdef OS2006
#include <hildon-widgets/hildon-program.h>
#else
#include <hildon/hildon-program.h>
#endif
#include <libosso.h>
#include "chordlib.h"
#include "prefs.h"

static HildonWindow *window_main;
static gboolean is_fullscreen;

static GtkComboBox *box_chord, *box_suffix;
static GtkToggleButton *(but_sub_chord[BASIC_KEY_CHORDS]);
static GtkDrawingArea *chord_area;

int cnt_chord, pos_chord, sub_chord;
int last_sub_chord = -1, last_box_suffix = -1;

PangoMatrix rotated_matrix = {
	.xy = -1,
	.yx = 1,
};

static void draw_one_chord(GtkWidget *widget, GdkGC *gc, PangoAttrList *alist, int x, int y, int w, int h, int fs,
			   unsigned chord, int start_bar)
{
	int w1, h1, r, xf, yf;
	int i, pos;
	PangoLayout *lay;
	PangoContext *ctx;
	gchar text[8];

	w1 = (w - 1) / (STRINGS_CNT - 1);
	gdk_gc_set_line_attributes(gc, 2, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
	for (i = 0; i < STRINGS_CNT; i++)
		gdk_draw_line(widget->window, gc, x + w1 * i, y, x + w1 * i, y + h - 1);
	h1 = (h - 2) / BAR_RANGE ;
	for (i = 1; i < BAR_RANGE + 1; i++)
		gdk_draw_line(widget->window, gc, x, y + h1 * i, x + w - 1, y + h1 * i);
	if (start_bar == 0)
		gdk_gc_set_line_attributes(gc, 6, GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER);
	gdk_draw_line(widget->window, gc, x - 1, y, x + w, y);

	r = (w1 - (w1 / 3)) / 2;
	xf = x - r;
	yf = y + h1 /2 - r;
	r *= 2;
	for (i = 0; i < STRINGS_CNT; i++) {
		pos = get_pos(chord, STRINGS_CNT - i - 1);
		if (pos) {
			pos -= start_bar + 1;
			gdk_draw_arc(widget->window, gc, TRUE, xf + i * w1, yf + pos * h1, r, r, 0, 360 * 64);
		}
	}

	xf = x + w + w1 / 2;
	yf = y + h1 / 2 - fs / 2;
	for (i = 0; i < BAR_RANGE; i++) {
		g_sprintf(text, "%d", i + start_bar + 1);
		lay = gtk_widget_create_pango_layout(widget, text);
		pango_layout_set_attributes(lay, alist);
		gdk_draw_layout(widget->window, gc, xf, yf + i * h1, lay);
		g_object_unref(lay);
	}

	xf = x - fs / 2;
	yf = y - h1 / 6;
	ctx = gtk_widget_create_pango_context(widget);
	pango_context_set_matrix(ctx, &rotated_matrix);
	for (i = 0; i < STRINGS_CNT; i++) {
		lay = pango_layout_new(ctx);
		pango_layout_set_text(lay, get_tone_name(STRINGS_CNT - i - 1, chord), -1);
		pango_layout_set_attributes(lay, alist);
		pango_layout_get_pixel_size(lay, &pos, NULL);
		gdk_draw_layout(widget->window, gc, xf + i * w1, yf - pos, lay);
		g_object_unref(lay);
	}
	g_object_unref(ctx);
}

static gboolean redraw_chord(GtkWidget *widget, GdkEventExpose *event, gpointer data)
{
	static GdkGC *chord_gc = NULL;
	PangoAttrList *alist;
	PangoAttribute *attr;
	int w, h, fs;
	int i;

	(void)event; (void)data; /* shut up, gcc */

	if (!chord_gc)
		chord_gc = gdk_gc_new(widget->window);

	w = widget->allocation.width / 4;
	h = widget->allocation.height;
	fs = h / 20;

	alist = pango_attr_list_new();
	attr = pango_attr_family_new("sans");
	attr->start_index = 0;
	attr->end_index = 1024;
	pango_attr_list_insert(alist, attr);
	attr = pango_attr_size_new_absolute(fs * PANGO_SCALE);
	attr->start_index = 0;
	attr->end_index = 1024;
	pango_attr_list_insert(alist, attr);

	h -= 40 + 2 * fs;
	for (i = 0; i < 4; i++) {
		if (pos_chord + i >= cnt_chord)
			break;
		draw_one_chord(widget, chord_gc, alist, 20 + w * i, 20 + 2 * fs,
			       w - 40 - 2 * fs, h, fs,
			       get_chord(pos_chord + i), get_chord_start_bar(pos_chord + i));
	}

	pango_attr_list_unref(alist);
	return TRUE;
}

static void recompute_chords(void)
{
	int act, i;

	act = gtk_combo_box_get_active(box_chord);

	for (i = 0; i < BASIC_KEY_CHORDS; i++)
		gtk_button_set_label(GTK_BUTTON(but_sub_chord[i]), get_basic_key_chord(act, i));

	if (sub_chord >= 0)
		generate_basic_key_chords(act, sub_chord);
	else
		generate_chords(act, gtk_combo_box_get_active(box_suffix));
	cnt_chord = get_chords_cnt();
	pos_chord = 0;
}

static void fill_combo_boxes(gboolean erase)
{
	int cnt, i, pos;

	cnt = get_chordnames_cnt();
	if (erase) {
		pos = gtk_combo_box_get_active(box_chord);
		for (i = cnt - 1; i >= 0; i--)
			gtk_combo_box_remove_text(box_chord, i);
	}
	for (i = 0; i < cnt; i++)
		gtk_combo_box_append_text(box_chord, chordnames[i]);
	if (erase)
		gtk_combo_box_set_active(box_chord, pos);

	cnt = get_chordsuff_cnt();
	if (erase) {
		pos = gtk_combo_box_get_active(box_suffix);
		for (i = cnt - 1; i >= 0; i--)
			gtk_combo_box_remove_text(box_suffix, i);
	}
	for (i = 0; i < cnt; i++)
		gtk_combo_box_append_text(box_suffix, chordsuff[i]);
	if (erase)
		gtk_combo_box_set_active(box_suffix, pos);
}

static void change_prefs(void)
{
	fill_combo_boxes(TRUE);
	recompute_chords();
	gtk_widget_queue_draw(GTK_WIDGET(chord_area));
}

static void change_chord(GtkComboBox *box, gpointer data)
{
	int pos;

	(void)data; /* shut up, gcc */

	if (box == box_suffix) {
		pos = gtk_combo_box_get_active(box_suffix);
		if (pos > 0)
			last_box_suffix = pos;
	}
	recompute_chords();
	gtk_widget_queue_draw(GTK_WIDGET(chord_area));
}

static void change_sub_chord(GtkToggleButton *but, gpointer data)
{
	int i;
	int idx = GPOINTER_TO_INT(data);

	if (gtk_toggle_button_get_active(but)) {
		sub_chord = last_sub_chord = idx;
		for (i = 0; i < BASIC_KEY_CHORDS; i++)
			if (i != idx && gtk_toggle_button_get_active(but_sub_chord[i]))
				gtk_toggle_button_set_active(but_sub_chord[i], FALSE);
		change_chord(NULL, NULL);
	} else
		if (sub_chord >= 0 && sub_chord == idx) {
			sub_chord = -1;
			change_chord(NULL, NULL);
		}
}

static void change_chord_pos(GtkButton *button, gpointer data)
{
	int i = GPOINTER_TO_INT(data);

	(void)button; /* shut up, gcc */
	if (i < 0) {
		pos_chord -= 4;
		if (pos_chord < 0)
			pos_chord = 0;
	} else {
		if (pos_chord + 4 < cnt_chord)
			pos_chord += 4;
	}
	gtk_widget_queue_draw(GTK_WIDGET(chord_area));
}

static void toggle_fullscreen(void)
{
	if (is_fullscreen)
		gtk_window_unfullscreen(GTK_WINDOW(window_main));
	else
		gtk_window_fullscreen(GTK_WINDOW(window_main));
	is_fullscreen = !is_fullscreen;
}

static gboolean key_pressed(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
	int action;
	int pos;

	(void)widget; (void)data; /* shut up, gcc */
	action = get_key_action(event->keyval);
	switch (action) {
	case -1:
		pos = -1;
		if (event->keyval >= GDK_a && event->keyval <= GDK_g) {
			pos = event->keyval - GDK_a;
			if (pos < 2)
				/* a - b */
				pos = pos * 2 + 9;
			else if (pos < 5)
				/* c - e */
				pos = pos * 2 - 4;
			else
				/* f - g */
				pos = pos * 2 - 5;
		} else if (event->keyval >= GDK_A && event->keyval <= GDK_G) {
			pos = event->keyval - GDK_A;
			if (pos < 2)
				/* A#, Bb */
				pos = 10;
			else if (pos < 4)
				/* C#, D# */
				pos = pos * 2 - 3;
			else if (pos == 4)
				/* Eb */
				pos = 3;
			else
				/* F#, G# */
				pos = pos * 2 - 4;
		}
		if (pos >= 0) {
			gtk_combo_box_set_active(box_chord, pos);
			return TRUE;
		}
		switch (event->keyval) {
		case GDK_m:
			/* mi */
			pos = 1;
			break;
		case GDK_u:
		case GDK_7:
			/* 7 */
			pos = 2;
			break;
		case GDK_M:
			/* mi7 */
			pos = 3;
			break;
		case GDK_U:
			/* 7maj */
			pos = 4;
			break;
		case GDK_space:
			/* dur */
			pos = 0;
			break;
		}
		if (pos >= 0) {
			gtk_combo_box_set_active(box_suffix, pos);
			return TRUE;
		}
		return FALSE;
	case A_NOTHING:
		return TRUE;
	case A_FULLSCREEN:
		toggle_fullscreen();
		return TRUE;
	case A_QUIT:
		gtk_main_quit();
		return TRUE;
	case A_PREV_SCR:
		change_chord_pos(NULL, GINT_TO_POINTER(-1));
		return TRUE;
	case A_NEXT_SCR:
		change_chord_pos(NULL, GINT_TO_POINTER(1));
		return TRUE;
	case A_PREV_CHORD:
		pos = gtk_combo_box_get_active(box_chord);
		gtk_combo_box_set_active(box_chord, pos > 0 ? pos - 1 : get_chordnames_cnt() - 1);
		return TRUE;
	case A_NEXT_CHORD:
		pos = gtk_combo_box_get_active(box_chord);
		gtk_combo_box_set_active(box_chord, pos < get_chordnames_cnt() - 1 ? pos + 1 : 0);
		return TRUE;
	case A_FIRST_SUBCHORD:
		if (sub_chord >= 0) {
			last_sub_chord = sub_chord;
			gtk_toggle_button_set_active(but_sub_chord[sub_chord], FALSE);
		} else if (last_sub_chord >= 0) {
			gtk_toggle_button_set_active(but_sub_chord[last_sub_chord], TRUE);
		}
		return TRUE;
	case A_PREV_SUBCHORD:
		if (sub_chord == 0)
			gtk_toggle_button_set_active(but_sub_chord[sub_chord], FALSE);
		else {
			if (sub_chord < 0)
				pos = BASIC_KEY_CHORDS - 1;
			else
				pos = sub_chord - 1;
			gtk_toggle_button_set_active(but_sub_chord[pos], TRUE);
		}
		return TRUE;
	case A_NEXT_SUBCHORD:
		if (sub_chord == BASIC_KEY_CHORDS - 1)
			gtk_toggle_button_set_active(but_sub_chord[sub_chord], FALSE);
		else {
			if (sub_chord < 0)
				pos = 0;
			else
				pos = sub_chord + 1;
			gtk_toggle_button_set_active(but_sub_chord[pos], TRUE);
		}
		return TRUE;
	case A_FIRST_SUFF:
		pos = gtk_combo_box_get_active(box_suffix);
		if (pos > 0) {
			last_box_suffix = pos;
			gtk_combo_box_set_active(box_suffix, 0);
		} else if (last_box_suffix >= 0)
			gtk_combo_box_set_active(box_suffix, last_box_suffix);
		return TRUE;
	case A_PREV_SUFF:
		pos = gtk_combo_box_get_active(box_suffix);
		gtk_combo_box_set_active(box_suffix, pos > 0 ? pos - 1 : get_chordsuff_cnt() - 1);
		return TRUE;
	case A_NEXT_SUFF:
		pos = gtk_combo_box_get_active(box_suffix);
		gtk_combo_box_set_active(box_suffix, pos < get_chordsuff_cnt() - 1 ? pos + 1 : 0);
		return TRUE;
	}
	return FALSE;
}

static gboolean tap_chord(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	GdkEventKey evkey;

	if (event->x < widget->allocation.width / 2)
		evkey.keyval = GDK_Page_Up;
	else
		evkey.keyval = GDK_Page_Down;
	return key_pressed(widget, &evkey, data);
}

static GtkWidget *main_layout(void)
{
	GtkVBox *vbox;
	GtkHBox *hbox, *hbox_tog;
#ifdef CHORD_POS_BUTTONS
	GtkButton *but;
	GtkImage *img;
#endif
	int i;

	vbox = GTK_VBOX(gtk_vbox_new(FALSE, 4));

	hbox = GTK_HBOX(gtk_hbox_new(FALSE, 0));

	/* combo boxes */
	box_chord = GTK_COMBO_BOX(gtk_combo_box_new_text());
	box_suffix = GTK_COMBO_BOX(gtk_combo_box_new_text());
	gtk_widget_set_size_request(GTK_WIDGET(box_chord), 80, -1);
	gtk_widget_set_size_request(GTK_WIDGET(box_suffix), 125, -1);
	fill_combo_boxes(FALSE);
	gtk_combo_box_set_active(box_chord, 0);
	gtk_combo_box_set_active(box_suffix, 0);
	g_signal_connect(G_OBJECT(box_chord), "changed", G_CALLBACK(change_chord), NULL);
	g_signal_connect(G_OBJECT(box_suffix), "changed", G_CALLBACK(change_chord), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(box_chord), FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(box_suffix), FALSE, FALSE, 0);

	/* chord buttons */
	hbox_tog = GTK_HBOX(gtk_hbox_new(TRUE, 0));
	for (i = 0; i < BASIC_KEY_CHORDS; i++) {
		but_sub_chord[i] = GTK_TOGGLE_BUTTON(gtk_toggle_button_new());
		g_object_set(G_OBJECT(but_sub_chord[i]), "can-focus", FALSE, NULL);
		g_signal_connect(G_OBJECT(but_sub_chord[i]), "toggled",
				 G_CALLBACK(change_sub_chord), GINT_TO_POINTER(i));
		gtk_box_pack_start_defaults(GTK_BOX(hbox_tog), GTK_WIDGET(but_sub_chord[i]));
	}
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(hbox_tog), TRUE, TRUE, 8);

	sub_chord = -1;
	recompute_chords();

#ifdef CHORD_POS_BUTTONS
	/* left/right buttons */
	img = GTK_IMAGE(gtk_image_new_from_stock(GTK_STOCK_GO_BACK, GTK_ICON_SIZE_BUTTON));
	g_object_set(G_OBJECT(img), "xpad", 6, NULL);
	but = GTK_BUTTON(gtk_button_new());
	gtk_container_add(GTK_CONTAINER(but), GTK_WIDGET(img));
	g_signal_connect(G_OBJECT(but), "clicked", G_CALLBACK(change_chord_pos), GINT_TO_POINTER(-1));
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(but), FALSE, FALSE, 0);

	img = GTK_IMAGE(gtk_image_new_from_stock(GTK_STOCK_GO_FORWARD, GTK_ICON_SIZE_BUTTON));
	g_object_set(G_OBJECT(img), "xpad", 6, NULL);
	but = GTK_BUTTON(gtk_button_new());
	gtk_container_add(GTK_CONTAINER(but), GTK_WIDGET(img));
	g_signal_connect(G_OBJECT(but), "clicked", G_CALLBACK(change_chord_pos), GINT_TO_POINTER(1));
	gtk_box_pack_start(GTK_BOX(hbox), GTK_WIDGET(but), FALSE, FALSE, 0);
#endif

	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(hbox), FALSE, FALSE, 0);

	/* chords */
	chord_area = GTK_DRAWING_AREA(gtk_drawing_area_new());
	gtk_widget_add_events(GTK_WIDGET(chord_area), GDK_BUTTON_PRESS_MASK);
	g_signal_connect(G_OBJECT(chord_area), "expose_event", G_CALLBACK(redraw_chord), NULL);
	g_signal_connect(G_OBJECT(chord_area), "button_press_event", G_CALLBACK(tap_chord), NULL);
	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(chord_area), TRUE, TRUE, 0);

	return GTK_WIDGET(vbox);
}

static void run_about(GtkWindow *parent)
{
	gtk_show_about_dialog(parent,
		"version", "0.5",
		"comments", "Licensed under GPLv2. Please send bug reports and feature requests to jbenc@upir.cz.",
		"copyright", "(c) 2007 Jiri Benc",
		"website", "http://upir.cz/maemo/");
}

static GtkMenu *main_menu(void)
{
	GtkMenuShell *menu;
	GtkWidget *item;

	menu = GTK_MENU_SHELL(gtk_menu_new());
	item = gtk_menu_item_new_with_label("Preferences");
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(run_prefs), window_main);
	gtk_menu_shell_append(menu, item);
	item = gtk_menu_item_new_with_label("Fullscreen");
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(toggle_fullscreen), window_main);
	gtk_menu_shell_append(menu, item);
	item = gtk_menu_item_new_with_label("About");
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(run_about), NULL);
	gtk_menu_shell_append(menu, item);
	item = gtk_menu_item_new_with_label("Exit");
	g_signal_connect(G_OBJECT(item), "activate", G_CALLBACK(gtk_main_quit), NULL);
	gtk_menu_shell_append(menu, item);
	return GTK_MENU(menu);
}

int main(int argc, char **argv)
{
	HildonProgram *program;

	gtk_init(&argc, &argv);
	program = hildon_program_get_instance();
	g_set_prgname("s4chords");
	g_set_application_name("4-string chords");
	osso_initialize("cz.upir.s4chords", "0.5", TRUE, NULL);
	window_main = HILDON_WINDOW(hildon_window_new());
	hildon_program_add_window(program, window_main);

	prefs_init(change_prefs);

	gtk_container_add(GTK_CONTAINER(window_main), main_layout());

	hildon_window_set_menu(window_main, main_menu());

	g_signal_connect(G_OBJECT(window_main), "key_press_event", G_CALLBACK(key_pressed), NULL);
	g_signal_connect(G_OBJECT(window_main), "delete_event", G_CALLBACK(gtk_main_quit), NULL);

	gtk_widget_show_all(GTK_WIDGET(window_main));
	gtk_main();
	return 0;
}
